/*-
 * Copyright 2005 Eric Anholt
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Authors:
 *    Eric Anholt <anholt@FreeBSD.org>
 */

/* Perlin noise functions, based on:
 * - http://www.devx.com/Intel/Article/20182
 * - http://freespace.virgin.net/hugo.elias/models/m_perlin.htm
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <liboil/liboil.h>

#include "pcg.h"
#include "pcg_int.h"

int use_cosine = 1;

void perlin_use_cosine(int use)
{
	use_cosine = use;
}

#if 0
#define COS_LEVELS	32

/* A cosine mapping of [0,1.0] to [0,1.0]: cos((1.0 - factor) * M_PI / 2); */
static float cosmap(float factor)
{
	static int initialized = 0;
	static float map[COS_LEVELS];

	if (factor <= 0.0)
		return 0.0;
	if (factor >= 1.0)
		return 1.0;

	if (!initialized) {
		int i;

		for (i = 0; i < COS_LEVELS - 1; i++) {
			float t = (float)i / (float)COS_LEVELS;
			map[i] = .5 + .5 * cos((1.0 - t) * M_PI);
		}
		/* Make the last value map to exactly 1 */
		map[COS_LEVELS - 1] = 1.0;

		initialized = 1;
	}

	return map[(int)((factor) * (COS_LEVELS - 1) + .5)];
}
#else
#define cosmap(_t) (.5 + .5 * cos((1.0 - _t) * M_PI))
#endif

void add_1d_perlin(float *data, int size, int offset, int stride,
    float weight)
{
	int stride_count = size / stride;
	float vals[stride_count + 1];
	float fracs[stride];
	float results[size];
	int i, x;

	PCG_ASSERT(offset < size, ("offset too large"));

	oil_init();

	/* Generate the random noise for this octave */
	for (i = 0; i < stride_count; i++) {
		vals[i] = pcg_random_f();
	}
	vals[i] = vals[0];	/* Add one wrap-around value */

	/* Prepare the fractions for smoothing */
	for (i = 0; i < stride; i++) {
		fracs[i] = (float)(stride - i) / stride;
		if (use_cosine)
			fracs[i] = cosmap(fracs[i]);
	}

	/* Generate the smoothed values for the random noise */
	x = 0;
	for (i = 0; i < stride_count; i++) {
		int t;

		for (t = 0; t < stride; t ++) {
			results[x] = vals[i] * fracs[t];
			results[x] += vals[i + 1] * (1.0 - fracs[t]);
			x++;
		}
	}
	oil_scalarmultiply_f32_ns(results, results, &weight, size);
	oil_add_f32(data + offset, data + offset, results, size - offset);
	oil_add_f32(data, data, results + (size - offset), offset);
}

float *create_1d_perlin(int size, float persistence, int levels)
{
	float *data;
	float weight, weight_factor;
	int stride = size / 2;
	int i;

	data = calloc(size, sizeof(float));
	if (data == NULL)
		return NULL;

	PCG_ASSERT(IS_POWER2(size), ("size is not a power of 2"));

	weight = persistence;
	weight_factor = 0.0;
	for (i = 0; i < levels; i++) {
		weight_factor += weight;
		weight *= persistence;
	}

	weight = persistence * (1.0 / weight_factor);

	for (i = 0; i < levels; i++) {
		int offset = pcg_random() % stride;

		add_1d_perlin(data, size, offset, stride, weight);
		weight *= persistence;
		stride /= 2;
	}

	return data;
}

void add_2d_perlin(float *data, int size, int ox, int oy, int stride,
    float weight)
{
	int stride_count = size / stride;
	float *vals[stride_count + 1];
	float fracs[stride];
	int i, y;

	PCG_ASSERT(offset < size, ("offset too large"));

	/* Prepare the rows of 1d perlin noise to add in */
	for (i = 0; i < stride_count; i++) {
		vals[i] = calloc(size, sizeof(float));
		add_1d_perlin(vals[i], size, ox, stride, 1.0);
	}
	vals[i] = vals[0];

	/* Prepare the fractions for smoothing */
	for (i = 0; i < stride; i++) {
		fracs[i] = (float)(stride - i) / stride;
		if (use_cosine)
			fracs[i] = cosmap(fracs[i]);
	}

	for (y = 0; y < size; y++) {
		int resulty = (y + oy) % size;
		float *datarow = data + resulty * size;
		float temp[size];
		float weight1 = fracs[y % stride] * weight;
		float weight2 = (1.0 - fracs[y % stride]) * weight;

		oil_scalarmultiply_f32_ns(temp, vals[y / stride], &weight1,
		    size);
		oil_add_f32(datarow, datarow, temp, size);
		oil_scalarmultiply_f32_ns(temp, vals[y / stride + 1], &weight2,
		    size);
		oil_add_f32(datarow, datarow, temp, size);
	}
	for (i = 0; i < stride_count; i++) {
		free(vals[i]);
	}
}

float *create_2d_perlin(int size, float persistence, int levels)
{
	float *data;
	float weight, weight_factor;
	int stride = size / 2;
	int i;

	data = calloc(size * size, sizeof(float));
	if (data == NULL)
		return NULL;

	PCG_ASSERT(IS_POWER2(size), ("size is not a power of 2"));

	weight = persistence;
	weight_factor = 0.0;
	for (i = 0; i < levels; i++) {
		weight_factor += weight;
		weight *= persistence;
	}

	weight = persistence * (1.0 / weight_factor);
	for (i = 0; i < levels; i++) {
		int offsetx = pcg_random() % stride;
		int offsety = pcg_random() % stride;

		add_2d_perlin(data, size, offsetx, offsety, stride, weight);
		stride /= 2;
		weight *= persistence;
	}

	return data;
}
