/*-
 * Copyright 2005 Eric Anholt
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Authors:
 *    Eric Anholt <anholt@FreeBSD.org>
 */

#include <stdio.h>
#include <stdlib.h>
#include <liboil/liboil.h>

#include "pcg.h"
#include "pcg_int.h"

/* Reaction-diffusion systems based on:
 * - "Generating Textures on Arbitrary Surfaces Using Reaction-Diffusion" by
 *   Greg Turk (University of North Carolina at Chapel Hill).
 * - "Models of Biological Pattern Formation" by Hans Meinhardt.  Academic
 *   Press, London: 1982.
 */

/* The calculate_laplacian function convolves src with the matrix:
 * 0  1  0
 * 1 -4  1
 * 0  1  0
 * and stores it in dst, with wrapping around w and h.
 */
static void calculate_laplacian(float *dst, float *src, int w, int h)
{
	int y;
	float srcmult = -4.0;
	float *dstrow = dst, *srcrow = src;

	oil_scalarmultiply_f32_ns(dst, src, &srcmult, w * h);

	for (y = 0; y < h; y++) {
		if (y > 0)
			oil_add_f32(dstrow, dstrow, srcrow - w, w);
		else
			oil_add_f32(dstrow, dstrow, src + (h - 1) * w, w);

		oil_add_f32(dstrow + 1, dstrow + 1, srcrow, w - 1);
		dstrow[0] += srcrow[w - 1];

		oil_add_f32(dstrow, dstrow, srcrow + 1, w - 1);
		dstrow[w - 1] += srcrow[0];

		if (y < h - 1)
			oil_add_f32(dstrow, dstrow, srcrow + w, w);
		else
			oil_add_f32(dstrow, dstrow, src + (h - 1) * w, w);

		dstrow += w;
		srcrow += w;
	}
}

struct turing_reaction *create_2d_reaction_turing(int w, int h, float Da,
    float Db, float s, float Bbase, float Bvary, float ainit, float binit)
{
	struct turing_reaction *react;
	int x, y;

	oil_init();

	react = calloc(sizeof(struct turing_reaction), 1);
	if (react == NULL) {
		return NULL;
	}

	react->a = malloc(w * h * sizeof(float));
	react->b = malloc(w * h * sizeof(float));
	react->B = malloc(w * h * sizeof(float));
	react->da = malloc(w * h * sizeof(float));
	react->db = malloc(w * h * sizeof(float));
	react->dda = malloc(w * h * sizeof(float));
	react->ddb = malloc(w * h * sizeof(float));
	if (react->a == NULL || react->b == NULL || react->B == NULL ||
	    react->da == NULL || react->db == NULL ||
	    react->dda == NULL || react->ddb == NULL) {
		free(react->a);
		free(react->b);
		free(react->B);
		free(react->da);
		free(react->db);
		free(react->dda);
		free(react->ddb);
		free(react);
		return NULL;
	}

	for (y = 0; y < h; y++) {
	    for (x = 0; x < w; x++) {
		react->a[y * w + x] = ainit;
		react->b[y * w + x] = binit;
		react->B[y * w + x] = Bbase +
		    (-1.0 + 2.0 * pcg_random_f()) * Bvary;
	    }
	}

	react->Da = Da;
	react->Db = Db;
	react->s = s;
	react->w = w;
	react->h = h;

	return react;
}

void iterate_2d_reaction_turing(struct turing_reaction *react, int iter)
{
	float *a = react->a;
	float *b = react->b;
	float *da = react->da;
	float *db = react->db;
	float *B = react->B;
	int w = react->w;
	int h = react->h;
	int x, y, i, ij;

	for (i = 0; i < iter; i++) {
		/* calculate change */
		calculate_laplacian(react->dda, react->a, w, h);
		calculate_laplacian(react->ddb, react->b, w, h);

		for (y = 0; y < h; y++) {
		    for (x = 0; x < w; x++) {
			float aij = a[y * w + x];
			float bij = b[y * w + x];

			da[y * w + x] =
			    react->s * (16.0 - aij * bij) +
			    react->Da * react->dda[y * w + x];
			db[y * w + x] =
			    react->s * (aij * bij - bij - B[y * w + x]) +
			    react->Db * react->ddb[y * w + x];
		    }
		}

		if (!react->locks) {
			oil_add_f32(a, a, da, w * h);
			oil_add_f32(b, b, db, w * h);
			for (ij = 0; ij < w * h; ij++) {
				if (a[ij] < 0.0)
					a[ij] = 0.0;
				if (b[ij] < 0.0)
					b[ij] = 0.0;
			}
		} else {
			/* Make change if the value is unlocked. */
			for (ij = 0; ij < w * h; ij++) {
				if (!react->locks[ij]) {
					a[ij] += da[ij];
					if (a[ij] < 0.0)
						a[ij] = 0.0;
					b[ij] += db[ij];
					if (b[ij] < 0.0)
						b[ij] = 0.0;
				}
			}
		}
		react->curiter++;
	}
}

void free_2d_reaction_turing(struct turing_reaction *react)
{
	free(react->a);
	free(react->b);
	free(react->B);
	free(react->da);
	free(react->db);
	free(react->dda);
	free(react->ddb);
	free(react->locks);
	free(react);
}

struct meinhardt_reaction *create_2d_reaction_meinhardt(int w, int h,
    float c, float Dg, float Ds, float Dr, float gamma,
    float ginit, float sinit, float svary, float rinit)
{
	struct meinhardt_reaction *react;
	int x, y;

	oil_init();

	react = calloc(sizeof(struct meinhardt_reaction), 1);
	if (react == NULL) {
		return NULL;
	}

	react->g1 = malloc(w * h * sizeof(float));
	react->g2 = malloc(w * h * sizeof(float));
	react->r = malloc(w * h * sizeof(float));
	react->s1 = malloc(w * h * sizeof(float));
	react->s2 = malloc(w * h * sizeof(float));
	react->dg1 = malloc(w * h * sizeof(float));
	react->dg2 = malloc(w * h * sizeof(float));
	react->dr = malloc(w * h * sizeof(float));
	react->ds1 = malloc(w * h * sizeof(float));
	react->ds2 = malloc(w * h * sizeof(float));
	react->ddg1 = malloc(w * h * sizeof(float));
	react->ddg2 = malloc(w * h * sizeof(float));
	react->ddr = malloc(w * h * sizeof(float));
	react->dds1 = malloc(w * h * sizeof(float));
	react->dds2 = malloc(w * h * sizeof(float));
	react->temp = malloc(w * h * sizeof(float));
	react->cs2g1g1 = malloc(w * h * sizeof(float));
	react->cs1g2g2 = malloc(w * h * sizeof(float));
	if (react->g1 == NULL || react->g2 == NULL || react->r == NULL ||
	    react->s1 == NULL || react->s2 == NULL ||
	    react->dg1 == NULL || react->dg2 == NULL || react->dr == NULL ||
	    react->ds1 == NULL || react->ds2 == NULL ||
	    react->ddg1 == NULL || react->ddg2 == NULL || react->ddr == NULL ||
	    react->dds1 == NULL || react->dds2 == NULL || react->temp == NULL ||
	    react->cs2g1g1 == NULL || react->cs1g2g2 == NULL) {
		free_2d_reaction_meinhardt(react);
		return NULL;
	}

	for (y = 0; y < h; y++) {
	    for (x = 0; x < w; x++) {
		react->g1[y * w + x] = ginit;
		react->g2[y * w + x] = ginit;
		react->r[y * w + x] = rinit;
		react->s1[y * w + x] = sinit +
		    (-1.0 + 2.0 * pcg_random_f()) * svary;
		react->s2[y * w + x] = sinit +
		    (-1.0 + 2.0 * pcg_random_f()) * svary;
	    }
	}

	react->Dg = Dg;
	react->Dr = Dr;
	react->Ds = Ds;
	react->c = c;
	react->gamma = gamma;
	react->w = w;
	react->h = h;

	/* Set alpha, beta to make the initial values stable, with the exception
	 * of the randomness
	 */
	react->A = c * sinit * ginit / rinit;
	react->B = 2 * c * sinit * ginit * ginit / rinit;

	fprintf(stderr, "args: Dg = %.04f, Dr = %.04f, Ds = %.04f\n",
	    react->Dg, react->Dr, react->Ds);
	fprintf(stderr, "args: A = %.04f, B = %.04f\n", react->A, react->B);
	fprintf(stderr, "args: c = %.04f, gamma = %.04f\n",
	    react->c, react->gamma);

	return react;
}

void iterate_2d_reaction_meinhardt(struct meinhardt_reaction *react, int iter)
{
	float *g1 = react->g1;
	float *g2 = react->g2;
	float *r = react->r;
	float *s1 = react->s1;
	float *s2 = react->s2;
	float *dg1 = react->dg1;
	float *dg2 = react->dg2;
	float *dr = react->dr;
	float *ds1 = react->ds1;
	float *ds2 = react->ds2;
	float *ddg1 = react->ddg1;
	float *ddg2 = react->ddg2;
	float *ddr = react->ddr;
	float *dds1 = react->dds1;
	float *dds2 = react->dds2;
	float *temp = react->temp;
	float *cs2g1g1 = react->cs2g1g1;
	float *cs1g2g2 = react->cs1g2g2;
	float negB = -(react->B);
	int w = react->w;
	int h = react->h;
	int i;

	for (i = 0; i < iter; i++) {
		float p0 = 0.001, p1 = 0.001;
		//printf("iter %d\n", i);
		/* calculate change */

		/* Calculate laplacians and apply diffusion factors */
		calculate_laplacian(ddg1, g1, w, h);
		oil_scalarmultiply_f32_ns(ddg1, ddg1, &react->Dg, w * h);
		calculate_laplacian(ddg2, g2, w, h);
		oil_scalarmultiply_f32_ns(ddg2, ddg2, &react->Dg, w * h);
		calculate_laplacian(ddr, r, w, h);
		/* Meinhardt marks the diffusion of r (that is, "Dr *
		 * ddr") optional.  Users can set Dr = 0 for this
		 * behavior.
		 */
		oil_scalarmultiply_f32_ns(ddr, ddr,  &react->Dr, w * h);
		calculate_laplacian(dds1, s1, w, h);
		oil_scalarmultiply_f32_ns(dds1, dds1, &react->Ds, w * h);
		calculate_laplacian(dds2, s2, w, h);
		oil_scalarmultiply_f32_ns(dds2, dds2, &react->Ds, w * h);

		/* Pre-calculate c*s2*g1*g1 */
		oil_multiply_f32(cs2g1g1, g1, g1, w * h);
		oil_multiply_f32(cs2g1g1, cs2g1g1, s2, w * h);
		oil_scalarmultiply_f32_ns(cs2g1g1, cs2g1g1, &react->c, w * h);

		/* Pre-calculate c*s1*g2*g2 */
		oil_multiply_f32(cs1g2g2, g2, g2, w * h);
		oil_multiply_f32(cs1g2g2, cs1g2g2, s1, w * h);
		oil_scalarmultiply_f32_ns(cs1g2g2, cs1g2g2, &react->c, w * h);

		/* dg1 = c*s2*g1*g1/r - A*g1 + ddg1 + p0 */
		oil_divide_f32(dg1, cs2g1g1, r, w * h); /* cs1g1g1 / r */
		oil_scalarmultiply_f32_ns(temp, g1, &react->A, w * h); /* =A*g1 */
		oil_subtract_f32(dg1, dg1, temp, w * h); /* -A*g1 */
		oil_add_f32(dg1, dg1, ddg1, w * h); /* + ddg1 */
		oil_scalaradd_f32_ns(dg1, dg1, &p0, w * h); /* + p0 */

		/* dg2 = c*s1*g2*g2/r - A*g2 + ddg2 + p0 */
		oil_divide_f32(dg2, cs1g2g2, r, w * h); /* cs1g2g2 / r */
		oil_scalarmultiply_f32_ns(temp, g2, &react->A, w * h); /* =A*g2 */
		oil_subtract_f32(dg2, dg2, temp, w * h); /* -A*g2 */
		oil_add_f32(dg2, dg2, ddg2, w * h); /* + ddg2 */
		oil_scalaradd_f32_ns(dg2, dg2, &p0, w * h); /* + p0 */

		/* r = c*s2*g1*g1 + c*s1*g2*g2 - B*r + ddr */
		oil_scalarmultiply_f32_ns(dr, r, &negB, w * h);
		oil_add_f32(dr, dr, cs2g1g1, w * h);
		oil_add_f32(dr, dr, cs1g2g2, w * h);
		oil_add_f32(dr, dr, ddr, w * h);

		/* ds1 = gamma * (g1 - s1) + dds1 + p1 */
		oil_subtract_f32(ds1, g1, s1, w * h);
		oil_scalarmultiply_f32_ns(ds1, ds1, &react->gamma, w * h);
		oil_add_f32(ds1, ds1, dds1, w * h);
		oil_scalaradd_f32_ns(ds1, ds1, &p1, w * h);

		/* ds2 = gamma * (g2 - s2) + dds2 + p1 */
		oil_subtract_f32(ds2, g2, s2, w * h);
		oil_scalarmultiply_f32_ns(ds2, ds2, &react->gamma, w * h);
		oil_add_f32(ds2, ds2, dds2, w * h);
		oil_scalaradd_f32_ns(ds2, ds2, &p1, w * h);

		/* Make change */
		if (!react->locks) {
			oil_add_f32(g1, g1, dg1, w * h);
			oil_add_f32(g2, g2, dg2, w * h);
			oil_add_f32(r, r, dr, w * h);
			oil_add_f32(s2, s2, ds2, w * h);
			oil_add_f32(s1, s1, ds1, w * h);
		} else {
			int ij;

			/* Make change if the value is unlocked. */
			for (ij = 0; ij < w * h; ij++) {
				if (react->locks[ij])
					continue;
				g1[ij] += dg1[ij];
				g2[ij] += dg2[ij];
				r[ij] += dr[ij];
				s1[ij] += ds1[ij];
				s2[ij] += ds2[ij];
			}
		}

		react->curiter++;
	}
}

void free_2d_reaction_meinhardt(struct meinhardt_reaction *react)
{
	free(react->g1);
	free(react->g2);
	free(react->r);
	free(react->s1);
	free(react->s2);
	free(react->dg1);
	free(react->dg2);
	free(react->dr);
	free(react->ds1);
	free(react->ds2);
	free(react->ddg1);
	free(react->ddg2);
	free(react->ddr);
	free(react->dds1);
	free(react->dds2);
	free(react->temp);
	free(react->cs2g1g1);
	free(react->cs1g2g2);
	free(react);
}
