/*-
 * Copyright 2005 Eric Anholt
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Authors:
 *    Eric Anholt <anholt@FreeBSD.org>
 */

/* Based on ideas from:
 * - http://www.devx.com/Intel/Article/20333/0
 */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include "GL/glut.h"
#include "pcg.h"

static GLuint Window = 0;
static int seed = 0;
static int size = 1024;
static int mainiter = 8;
static GLfloat *heightmap;
static GLint *colorindices;
static GLfloat *vertices, *normals;
static GLuint CloudTex;

/* Map x,y indices of heightmap to [-1.0,1.0]^2 */
#define YOFF(_y) (-1.0 + 2.0 * (GLfloat)(_y) / (GLfloat)size)
#define XOFF(_x) (-1.0 + 2.0 * (GLfloat)(_x) / (GLfloat)size)
#define ZOFF(_x, _y) (heightmap[(_y) * size + (_x)])

/* Difference between adjacent points, for normal calculation */
#define XOFF_INCR	(1.0 / size)
#define YOFF_INCR	(1.0 / size)

static void make_normal(GLfloat *d0, GLfloat *d1, GLfloat *norm)
{
	GLfloat len;

	norm[0] = d0[1] * d1[2] - d0[2] * d1[1];
	norm[1] = d0[2] * d1[0] - d0[0] * d1[2];
	norm[2] = d0[0] * d1[1] - d0[1] * d1[0];
	len = sqrt(norm[0] * norm[0] + norm[1] * norm[1] + norm[2] * norm[2]);
	norm[0] /= len;
	norm[1] /= len;
	norm[2] /= len;
}

static void make_vertices_from_heightmap()
{
	int x, y;

	for (y = 0; y < size; y++) {
	    for (x = 0; x < size; x++) {
		GLint i = y * size + x;

		vertices[i * 3 + 0] = XOFF(x);
		vertices[i * 3 + 1] = YOFF(y);
		vertices[i * 3 + 2] = ZOFF(x, y);
	    }
	}
}

static void make_normals_for_vertices()
{
	int x, y;

	for (y = 0; y < size; y++) {
	    GLint yp = (y + size - 1) % size;
	    GLint yn = (y + 1) % size;
	    for (x = 0; x < size; x++) {
		int xp = (x + size - 1) % size;
		int xn = (x + 1) % size;
		GLfloat v0[3], v1[3];

		v0[0] = 2.0 * XOFF_INCR;
		v0[1] = 0.0;
		v0[2] = vertices[(y * size + xn) * 3 + 2] -
			vertices[(y * size + xp) * 3 + 2];

		v1[0] = 0.0;
		v1[1] = 2.0 * YOFF_INCR;
		v1[2] = vertices[(yn * size + x) * 3 + 2] -
			vertices[(yp * size + x) * 3 + 2];

		make_normal(v0, v1, &normals[(y * size + x) * 3]);
	    }
	}
}

static void choose_vertex_colors(GLfloat *sandheight, GLfloat *snowcaps)
{
	int i;

	for (i = 0; i < size * size; i++) {
		if (heightmap[i] > .15 + snowcaps[i] * .20)
			colorindices[i] = 2;	/* Snow */
		else if (heightmap[i] > sandheight[i] * .025)
			colorindices[i] = 0;	/* Ground */
		else
			colorindices[i] = 1;	/* Sand */
	}
}

#define CLOUD_SIZE	2048
#define CLOUD_COVER	.4	/* Larger = fewer clouds */
#define CLOUD_SHARPNESS	.2	/* Larger = fuzzier clouds */

static void create_clouds()
{
	GLfloat *cloud_noise;
	GLfloat *tex;
	static int first_time = 1;
	int i;

	cloud_noise = create_2d_perlin(CLOUD_SIZE, .8, 8);
	tex = malloc(CLOUD_SIZE * CLOUD_SIZE * 4 * sizeof(GLfloat));

	for (i = 0; i < CLOUD_SIZE * CLOUD_SIZE; i++) {
		cloud_noise[i] -= CLOUD_COVER;
		if (cloud_noise[i] < 0.0)
			cloud_noise[i] = 0.0;
		cloud_noise[i] = 1.0 - powf(CLOUD_SHARPNESS, cloud_noise[i]);

		tex[i * 4 + 0] = 1.0;
		tex[i * 4 + 1] = 1.0;
		tex[i * 4 + 2] = 1.0;
		tex[i * 4 + 3] = cloud_noise[i];
	}

	if (first_time) {
		glGenTextures(1, &CloudTex);
		first_time = 0;
	}
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, CloudTex);
	glTexImage2D(GL_TEXTURE_2D, 0, 4, CLOUD_SIZE, CLOUD_SIZE, 0,
	    GL_RGBA, GL_FLOAT, tex);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
	glDisable(GL_TEXTURE_2D);
	free(tex);
}

static void init_heightmap_seeded(int seed)
{
	GLfloat *amp;
	GLfloat *noise;
	GLfloat *exponent;
	GLfloat *sandheight, *snowcaps;
	GLint x, y;

	if (heightmap == NULL) {
		heightmap = malloc(size * size * sizeof(GLfloat *));
		if (heightmap == NULL)
			errx(1, "malloc");
	}
	pcg_random_seed(seed);

	amp = create_2d_perlin(size, .25, mainiter);
	noise = create_2d_perlin(size, .5, mainiter);
	/* Low persistence on exponent, to make the transition from smooth to
	 * craggy slower.
	 */
	exponent = create_2d_perlin(size, .125, mainiter / 2);
	/* Low persistence on sand/snow height, to make the sand level change
	 * more slowly.
	 */
	sandheight = create_2d_perlin(size, .25, mainiter / 2);
	snowcaps = create_2d_perlin(size, .5, mainiter / 2);

	for (y = 0; y < size; y++) {
	    for (x = 0; x < size; x++) {
		GLint i = y * size + x;
		GLfloat noisemap, expmap;

		/* Map noise to [-1.0,1.0] */
		noisemap = noise[i] * 2.0 - 1.0;
		/* Map exponent to [0.0, 2.0] */
		expmap = exponent[i] * 2.0;
		/* Take noise ^ exp, for varying smoothing */
		heightmap[i] = powf(fabs(noisemap), expmap);
		if (noisemap < 0)
			heightmap[i] = -heightmap[i];
		heightmap[i] *= amp[i];
	    }
	}
	free(amp);
	free(noise);
	free(exponent);

	/* Now, prepare the data that we'll want in the GL calls later */
	if (vertices == NULL) {
		vertices = malloc(size * size * sizeof(GLfloat) * 3);
		normals = malloc(size * size * sizeof(GLfloat) * 3);
		colorindices = malloc(size * size * sizeof(GLint));
	}
	make_vertices_from_heightmap();
	make_normals_for_vertices();
	choose_vertex_colors(sandheight, snowcaps);
	create_clouds();

	free(sandheight);
	free(snowcaps);
}

static void draw_waterline(void)
{
	GLfloat water[4] = {0.0, 0.0, 0.7, .8};

	/* Draw a water line */
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, water);
	glNormal3f(0.0, 0.0, 1.0);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glBegin(GL_POLYGON);
		glVertex3f(-100.0, -100.0, 0.0);
		glVertex3f(-100.0,  100.0, 0.0);
		glVertex3f( 100.0,  100.0, 0.0);
		glVertex3f( 100.0, -100.0, 0.0);
	glEnd();
	glDisable(GL_BLEND);
}

#define SKYBOX_W	50.0
#define SKYBOX_H	5.0
static void draw_sky(void)
{
	glColor4f(0.2, 0.2, 1.0, 1.0);
	glDisable(GL_LIGHTING);

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, CloudTex);
	/* Top of skybox, with clouds */
	glBegin(GL_POLYGON);
		glTexCoord2f( 0.0, 0.0 );
		glVertex3f( -SKYBOX_W, -SKYBOX_W, SKYBOX_H );
		glTexCoord2f( 1.0, 0.0 );
		glVertex3f(  SKYBOX_W, -SKYBOX_W, SKYBOX_H );
		glTexCoord2f( 1.0, 1.0 );
		glVertex3f(  SKYBOX_W,  SKYBOX_W, SKYBOX_H );
		glTexCoord2f( 0.0, 1.0 );
		glVertex3f( -SKYBOX_W,  SKYBOX_W, SKYBOX_H );
	glEnd();
	glDisable(GL_TEXTURE_2D);

	/* Sides of skybox */
	glBegin(GL_QUADS);
		glVertex3f(  SKYBOX_W,  SKYBOX_W, SKYBOX_H );
		glVertex3f(  SKYBOX_W, -SKYBOX_W, SKYBOX_H );
		glVertex3f(  SKYBOX_W, -SKYBOX_W, -SKYBOX_H );
		glVertex3f(  SKYBOX_W,  SKYBOX_W, -SKYBOX_H );

		glVertex3f(  -SKYBOX_W,  SKYBOX_W, SKYBOX_H );
		glVertex3f(  -SKYBOX_W, -SKYBOX_W, SKYBOX_H );
		glVertex3f(  -SKYBOX_W, -SKYBOX_W, -SKYBOX_H );
		glVertex3f(  -SKYBOX_W,  SKYBOX_W, -SKYBOX_H );

		glVertex3f(  SKYBOX_W, -SKYBOX_W, SKYBOX_H );
		glVertex3f( -SKYBOX_W, -SKYBOX_W, SKYBOX_H );
		glVertex3f( -SKYBOX_W, -SKYBOX_W, -SKYBOX_H );
		glVertex3f(  SKYBOX_W, -SKYBOX_W, -SKYBOX_H );

		glVertex3f(  SKYBOX_W,  SKYBOX_W, SKYBOX_H );
		glVertex3f( -SKYBOX_W,  SKYBOX_W, SKYBOX_H );
		glVertex3f( -SKYBOX_W,  SKYBOX_W, -SKYBOX_H );
		glVertex3f(  SKYBOX_W,  SKYBOX_W, -SKYBOX_H );
	glEnd();

	glEnable(GL_LIGHTING);
}

static void draw(void)
{
	GLint x, y;
	GLfloat colors[][4] = {
		{0.7, 0.5, 0.0, 1.0},	/* ground */
		{1.0, 0.8, 0.2, 1.0},	/* sand */
		{1.0, 1.0, 1.0, 1.0},	/* snow */
	};

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

	for (y = 0; y < size - 1; y++) {
	    GLint yn = (y + 1) % size;
	    for (x = 0; x < size - 1; x++) {
		GLint xn = (x + 1) % size;

		glBegin(GL_POLYGON);
			glColor4fv(colors[colorindices[y * size + x]]);
			glNormal3fv(&normals [(y  * size + x ) * 3]);
			glVertex3fv(&vertices[(y  * size + x ) * 3]);
			glColor4fv(colors[colorindices[yn * size + x]]);
			glNormal3fv(&normals [(yn * size + x ) * 3]);
			glVertex3fv(&vertices[(yn * size + x ) * 3]);
			glColor4fv(colors[colorindices[yn * size + xn]]);
			glNormal3fv(&normals [(yn * size + xn) * 3]);
			glVertex3fv(&vertices[(yn * size + xn) * 3]);
			glColor4fv(colors[colorindices[y * size + xn]]);
			glNormal3fv(&normals [(y  * size + xn) * 3]);
			glVertex3fv(&vertices[(y  * size + xn) * 3]);
		glEnd();
	    }
	}

	glDisable(GL_COLOR_MATERIAL);

	draw_waterline();
	draw_sky();

	glPopMatrix();

	glutSwapBuffers();
}

/* change view Angle, exit upon ESC */
static void key(unsigned char k, int x, int y)
{
	(void) x;
	(void) y;

	switch (k) {
	case 27:
		exit(0);
	default:
		init_heightmap_seeded(++seed);
		draw();
	}
}



/* new window size or exposure */
static void reshape(int width, int height)
{
	GLfloat ar = (GLfloat)width / (GLfloat)height;

	glViewport(0, 0, (GLint)width, (GLint)height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(90, ar, .25, 100.0);
	gluLookAt(
	    .75, .75, 0.4,	/* eye position */
	    0.0, 0.0, 0.2,	/* center */
	    0.0, 0.0, 1.0	/* up */
	    );
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	//glTranslatef( 0.0, 0.0, 1.0 );
}


static void init(void)
{
	GLfloat white[4] = {1.0, 1.0, 1.0, 1.0};
	GLfloat lightpos[4] = {-1.0, 1.0, 3.0};
	GLfloat ambi[4] = {0.2, 0.2, 0.2, 1.0};
	//GLfloat diff[4] = {0.8, 0.8, 0.8, 0.8};
	GLfloat diff[4] = {1.0, 1.0, 1.0, 1.0};

	glDisable(GL_DITHER);
	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);

	glEnable(GL_LIGHTING);
	glLightfv(GL_LIGHT0, GL_POSITION, lightpos);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambi);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diff);
	glEnable(GL_LIGHT0);

	//glClearColor(1.0, 1.0, 1.0, 1.0);
	glEnable(GL_FOG);
	glFogfv(GL_FOG_COLOR, white);
#if 0
	glFogi(GL_FOG_MODE, GL_EXP2);
	glFogf(GL_FOG_DENSITY, .2);
#else
	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogf(GL_FOG_START, 0.0);
	glFogf(GL_FOG_END, 20.0);
#endif
	glHint(GL_FOG_HINT, GL_NICEST);

	init_heightmap_seeded(seed);

	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
}

int main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	glutInitWindowPosition(0, 0);
	glutInitWindowSize(640, 480);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);

	Window = glutCreateWindow("terrain map");
	if (!Window) {
		exit(1);
	}

	init();

	glutReshapeFunc(reshape);
	glutKeyboardFunc(key);
	glutDisplayFunc(draw);
	glutMainLoop();

	return 0;
}
