/*-
 * Copyright 2005 Eric Anholt
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Authors:
 *    Eric Anholt <anholt@FreeBSD.org>
 */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "GL/glut.h"
#include "pcg.h"

static GLuint Window = 0;
static GLuint TexObj[2];
static GLfloat Angle = 0.0f;
static int seed = 0;
static int first_time = 1;
enum states {
	TURING,
	MEINHARDT,
	TURINGTURING,
	MEINHARDTMEINHARDT,
};

static GLboolean curstate = MEINHARDT;

static void draw(void);

/* Map the values in tex from [min,max] to [0.0, 1.0] */
static void normalize_tex(GLfloat *dst, GLfloat *src, int len)
{
	GLfloat min = src[0], max=src[0];
	int i;

	for (i = 1; i < len; i++) {
		if (src[i] < min)
			min = src[i];
		if (src[i] > max)
			max = src[i];
	}
	printf("min = %f, max=%f\n", min, max);
	for (i = 0; i < len; i++) {
		dst[i] = (src[i] - min) / (max - min);
	}
}

static void load_tex_unnormalized(float *in, int w, int h)
{
	float *vals;

	vals = malloc(w * h * sizeof(float));
	if (vals == NULL)
		return;

	normalize_tex(vals, in, w * h);
	glTexImage2D(GL_TEXTURE_2D, 0, 1, w, h, 0, GL_LUMINANCE,
	    GL_FLOAT, vals);

	free(vals);
}

static void iterate_and_draw_turing(struct turing_reaction *react, int maxiter,
    int steps)
{
	int i;

	for (i = 1; i <= steps; i++) {
		iterate_2d_reaction_turing(react, maxiter / steps);
		printf("At iteration %d\n", react->curiter);
		load_tex_unnormalized(react->b, react->w, react->h);
		draw();
	}
}

static void iterate_and_draw_meinhardt(struct meinhardt_reaction *react,
    int maxiter, int steps)
{
	int i;

	for (i = 1; i <= steps; i++) {
		iterate_2d_reaction_meinhardt(react, maxiter / steps);
		printf("At iteration %d\n", react->curiter);
		load_tex_unnormalized(react->g1, react->w, react->h);
		draw();
	}
}

#define W 128
#define H 128
static void init_tex_seeded(int seed)
{
	pcg_random_seed(seed);

	if (first_time)
		first_time = 0;

	if (curstate == TURING) {
		struct turing_reaction *react;

		react = create_2d_reaction_turing(W, H, .125, .03125, .01,
		    12, .1, 4, 4);

		iterate_and_draw_turing(react, 5000, 16);

		free_2d_reaction_turing(react);
	} else if (curstate == TURINGTURING) {
		struct turing_reaction *react;
		int ij;

		/* First, set up the reaction at a fast pace, and iterate for a
		 * while.
		 */
		react = create_2d_reaction_turing(W, H, .125, .03125, .01,
		    12, .1, 4, 4);

		iterate_and_draw_turing(react, 5000, 16);

		/* Now, lock down the activated sections */
		react->locks = calloc(W, H);
		for (ij = 0; ij < W * H; ij++) {
			if (react->b[ij] < .2)
				react->locks[ij] = 1;
		}
		/* And speed up the reaction */
		react->s *= 3;

		iterate_and_draw_turing(react, 3000, 16);

		free_2d_reaction_turing(react);
	} else if (curstate == MEINHARDT) {
		struct meinhardt_reaction *react;

		/* w, h, c, Dg, Ds, Dr, gamma, ginit, sinit, svary, rinit, iter */
		react = create_2d_reaction_meinhardt(W, H, .002, .016, .08, 0.0, .01,
		    .75, .75, .2, .14);

		iterate_and_draw_meinhardt(react, 16000, 16000/25);

		free_2d_reaction_meinhardt(react);
	} else if (curstate == MEINHARDTMEINHARDT) {
		struct meinhardt_reaction *react;
		int ij;

		/* w, h, c, Dg, Ds, Dr, gamma, ginit, sinit, svary, rinit, iter */
		react = create_2d_reaction_meinhardt(W, H, .002, .016, .08, 0.0, .01,
		    .75, .75, .2, .14);

		iterate_and_draw_meinhardt(react, 16000, 16000/25);

		/* Now, lock down the activated sections */
		react->locks = calloc(W, H);
		for (ij = 0; ij < W * H; ij++) {
			if (react->g1[ij] > 1.0)
				react->locks[ij] = 1;
		}
		/* And slow down the diffusion */
		react->Dg /= 5;

		iterate_and_draw_meinhardt(react, 16000, 16000/25);

		free_2d_reaction_meinhardt(react);
	}
}

static void draw(void)
{
	if (first_time) {
		init_tex_seeded(seed);
	}

	glClear(GL_COLOR_BUFFER_BIT);

	glColor3f(1.0, 1.0, 1.0);

	/* draw first polygon */
	glPushMatrix();
	glTranslatef(-1.0, 0.0, 0.0);
	glRotatef(Angle, 0.0, 0.0, 1.0);
	glBindTexture(GL_TEXTURE_2D, TexObj[0]);
	glBegin(GL_POLYGON);
	glTexCoord2f(0.0, 0.0);   glVertex2f(-1.0, -1.0);
	glTexCoord2f(1.0, 0.0);   glVertex2f( 1.0, -1.0);
	glTexCoord2f(1.0, 1.0);   glVertex2f( 1.0,  1.0);
	glTexCoord2f(0.0, 1.0);   glVertex2f(-1.0,  1.0);
	glEnd();
	glPopMatrix();

	glutSwapBuffers();
}

/* change view Angle, exit upon ESC */
static void key(unsigned char k, int x, int y)
{
	(void) x;
	(void) y;

	switch (k) {
	case 27:
		exit(0);
	case 't':
		if (curstate == TURING)
			curstate = TURINGTURING;
		else
			curstate = TURING;
		init_tex_seeded(seed);
		draw();
		break;
	case 'm':
		if (curstate == MEINHARDT)
			curstate = MEINHARDTMEINHARDT;
		else
			curstate = MEINHARDT;
		init_tex_seeded(seed);
		draw();
		break;
	default:
		init_tex_seeded(++seed);
		draw();
	}
}



/* new window size or exposure */
static void reshape(int width, int height)
{
	glViewport(0, 0, (GLint)width, (GLint)height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-2.0, 2.0, -2.0, 2.0, 6.0, 20.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -8.0);
}


static void init(void)
{
	glDisable(GL_DITHER);

	/* Setup texturing */
	glEnable(GL_TEXTURE_2D);

	/* generate texture object IDs */
	glGenTextures(2, TexObj);

	/* setup first texture object */
	glBindTexture(GL_TEXTURE_2D, TexObj[0]);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
}

int main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	glutInitWindowPosition(0, 0);
	glutInitWindowSize(300, 300);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);

	Window = glutCreateWindow("2d noise texture");
	if (!Window) {
		exit(1);
	}

	init();

	glutReshapeFunc(reshape);
	glutKeyboardFunc(key);
	glutDisplayFunc(draw);
	glutMainLoop();

	return 0;
}
